from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from authentikate.base_models import AuthentikateSettings
from typing import Optional
from pydantic import ValidationError

cached_settings: Optional[AuthentikateSettings] = None


def prepare_settings() -> AuthentikateSettings:
    """Prepare the settings

    Prepare the settings for authentikate from django_settings.
    This function will raise a ImproperlyConfigured exception if the settings are
    not correct.

    Returns
    -------
    AuthentikateSettings
        The settings

    Raises
    ------
    ImproperlyConfigured
        When the settings are not correct
    """

    try:
        group = settings.AUTHENTIKATE
    except AttributeError:
        raise ImproperlyConfigured("Missing setting AUTHENTIKATE")

    try:

        return AuthentikateSettings(
            **group
        )

    except ValidationError as e:
        raise ImproperlyConfigured(
            "Invalid settings for AUTHENTIKATE. Please check your settings."
        ) from e


def get_settings() -> AuthentikateSettings:
    """Get the settings

    Returns
    -------

    AuthentikateSettings
        The settings
    """
    global cached_settings
    if not cached_settings:
        cached_settings = prepare_settings()
    return cached_settings



from django.core.exceptions import PermissionDenied


class AuthentikateError(Exception):
    """Base class for all authentikate errors that are
    not permission related. Inherits from Exception"""

    pass


class AuthentikatePermissionDenied(PermissionDenied):
    """Base class for all authentikate permission errors. Inherits from
    django.core.exceptions.PermissionDenied"""

    pass


class AuthentikateTokenExpired(AuthentikatePermissionDenied):
    """Raised when a token is expired"""

    pass


class JwtTokenError(AuthentikatePermissionDenied):
    """Base class for all JWT token errors"""

    pass


class MalformedJwtTokenError(JwtTokenError):
    """Raised when a token is malformed."""

    pass


class InvalidJwtTokenError(JwtTokenError):
    """Raised when a token is invalid."""

    pass


class AuthentikateUserNotFound(AuthentikatePermissionDenied):
    """Raised when a user is not found"""

    pass



from typing import Protocol




class UserModel(Protocol):
    """A protocol for the user model"""
    id: int
    """The id of the user"""
    
    
    
    
class ClientModel(Protocol):
    """A protocol for the client model"""

    id: int
    """The id of the client"""

    client_id: str
    """The name of the client"""
    
    
    
    def has_scope(self, scope: str) -> bool:
        """Check if the user has the given scope"""
        ...
    


class OrganizationModel(Protocol):
    """A protocol for the organizaition model"""

    id: int
    """The id of the client"""

    slug: str
    """The name of orgnaization"""


from django.contrib.auth.models import Group
from authentikate import base_models, models
import logging


logger = logging.getLogger(__name__)


def token_to_username(token: base_models.JWTToken) -> str:
    """Convert a JWT token to a username

    Parameters
    ----------
    token : structs.JWTToken
        The token to convert

    Returns
    -------

    str
        The username



    """
    return f"{token.iss}_{token.sub}"





async def aset_user_groups(user: models.User, roles: list[str]) -> None:
    """Add a list of roles to a user

    Roles are added as groups

    Parameters
    ----------
    user : models.User
        The user to add the roles to
    roles : list[str]
        The roles to add
    """
    for role in roles:
        g, _ = await Group.objects.aget_or_create(name=role)
        await user.groups.aadd(g)


def set_user_groups(user: models.User, roles: list[str]) -> None:
    """Add a list of roles to a user

    Roles are added as groups

    Parameters
    ----------
    user : models.User
        The user to add the roles to
    roles : list[str]
        The roles to add
    """
    for role in roles:
        g, _ = Group.objects.get_or_create(name=role)
        user.groups.add(g)


async def aexpand_organization_from_token(
    token: base_models.JWTToken,
) -> models.Organization:
    """
    Expand an organization from the provided JWT token.
    """
    org, _ = await models.Organization.objects.aget_or_create(
        slug=token.active_org
    )
    return org


async def aexpand_user_from_token(
    token: base_models.JWTToken,
) -> models.User:
    """
    Expand a user from the provided JWT token.
    """
    
    try:
        user = await models.User.objects.aget(sub=token.sub, iss=token.iss)
        if user.changed_hash != token.changed_hash:
            # User has changed, update the user object
            user.first_name = token.preferred_username
            user.changed_hash = token.changed_hash
            
            if token.active_org:
                current_org, _ = await models.Organization.objects.aget_or_create(
                    slug=token.active_org or "",
                )
                
                user.active_organization = current_org
                
            
            
            await user.asave()
            await aset_user_groups(user, token.roles)
            
            
        return user

    except models.User.DoesNotExist:

        user = models.User(
            sub=token.sub,
            username=token_to_username(token),
            iss=token.iss,
            first_name=token.preferred_username,
        )
        user.set_unusable_password()
        user.first_name = token.preferred_username
        user.changed_hash = token.changed_hash
        
        if token.active_org:
            current_org, _ = await models.Organization.objects.aget_or_create(
                slug=token.active_org or "",
            )
            
            user.active_organization = current_org
        
        await user.asave()
        await aset_user_groups(user, token.roles)
        return user
   
def expand_user_from_token(
    token: base_models.JWTToken,
) -> models.User:
    """
    Expand a user from the provided JWT token.
    """
    
    try:
        user = models.User.objects.get(sub=token.sub, iss=token.iss)
        if user.changed_hash != token.changed_hash:
            # User has changed, update the user object
            user.first_name = token.preferred_username
            user.changed_hash = token.changed_hash
            set_user_groups(user, token.roles)
            
            if token.active_org:
                current_org, _ = models.Organization.objects.get_or_create(
                    identifier=token.active_org)
                
                user.active_organization = current_org
            
            user.save()
            
        return user

    except models.User.DoesNotExist:
        preexisting_user = models.User.objects.filter(
            username=token.preferred_username
        ).first()

        user = models.User(
            sub=token.sub,
            username=token_to_username(token)
            if preexisting_user
            else token.preferred_username,
            iss=token.iss,
            first_name=token.preferred_username,
        )
        user.set_unusable_password()
        user.first_name = token.preferred_username
        user.changed_hash = token.changed_hash
        
        if token.active_org:
            current_org, _ = models.Organization.objects.get_or_create(
                identifier=token.active_org)
            
            user.active_organization = current_org
        
        
        user.save()
        set_user_groups(user, token.roles)
        return user 
    
async def aexpand_client_from_token(
    token: base_models.JWTToken,
) -> models.Client:
    """
    Expand a client from the provided JWT token.
    """
    client, _ = await models.Client.objects.aget_or_create(
        client_id=token.client_id, iss=token.iss
    )
    return client


def expand_client_from_token(
    token: base_models.JWTToken,
) -> models.Client:
    """
    Expand a client from the provided JWT token.
    """
    client, _ = models.Client.objects.get_or_create(
        client_id=token.client_id, iss=token.iss
    )
    return client






from joserfc import jwt
from authentikate import base_models, errors


def decode_token(
    token: str,  settings: base_models.AuthentikateSettings
) -> base_models.JWTToken:
    """Decode a JWT token

    Parameters
    ----------
    token : str
        The token to decode
    algorithms : list
        The algorithms to use to decode the token
    public_key : str
        The public key to use to decode the token

    Returns
    -------
    structs.JWTToken
        The decoded token
    """
    try:
        decoded = jwt.decode(token, settings.load_key)
    except Exception as e:
        raise errors.InvalidJwtTokenError("Error decoding token") from e

    try:
        return base_models.JWTToken(**{"raw": token, **decoded.claims})
    except TypeError as e:
        raise errors.MalformedJwtTokenError("Error decoding token") from e



import strawberry
from typing import Awaitable, Callable, Any, Optional, List, Union
from graphql import GraphQLError
from strawberry.extensions.field_extension import AsyncExtensionResolver
from strawberry.schema_directive import Location
from kante.types import Info
from strawberry.extensions import FieldExtension
from strawberry.schema_directive import Location
from strawberry.types.field import StrawberryField
from authentikate.base_models import JWTToken
from authentikate.protocols import UserModel, ClientModel
from typing import cast


@strawberry.schema_directive(locations=[Location.FIELD_DEFINITION])
class Auth:
    """ A directive to enforce authentication and authorization on fields."""
    required_scopes: Optional[List[str]] = strawberry.directive_field(name="required_scopes", default=None)
    required_roles: Optional[List[str]] = strawberry.directive_field(name="required_roles", default=None)

    

class AuthExtension(FieldExtension):
    
    
    def __init__(self, scopes: Optional[List[str]] | str = None, roles: Optional[List[str]] = None) -> None:
        """Initialize the AuthExtension with optional scopes and roles."""
        if isinstance(scopes, str):
            scopes = [scopes]
        if roles and isinstance(roles, str):
            roles = [roles]
        
        self.scopes: Optional[List[str]] = scopes
        self.roles: Optional[List[str]] = roles
        
        
        
    def apply(self, field: StrawberryField) -> None:
        """Apply the Auth directive to the field.

        Args:
            field (StrawberryField): The authentication field to which the directive will be applied.
        """
        assert not field.is_subscription, "Auth directive cannot be applied to subscriptions use AuthSubscribeExtension instead."
        field.directives.append(Auth(required_scopes=self.scopes, required_roles=self.roles))

    def resolve(
        self, next_: Callable[..., Any], source: Any, info: Info, **kwargs
    ) -> Any:
        """ Resolve the field with authentication checks."""
        if not info.context.request.user:
            raise GraphQLError("Authentication required")
        
        
        try:
            token: JWTToken = info.context.request.get_extension("token")
            
            if self.scopes and not token.has_scopes(self.scopes):
                raise GraphQLError(f"User does not have the required scopes: {self.scopes}")
        
            if self.roles and not token.has_roles(self.roles):
                raise GraphQLError(f"User does not have the required roles: {', '.join(self.roles)}")
            
            
        except KeyError:
            raise GraphQLError("Token not found in request context")
        
        
        
        return next_(source, info, **kwargs)
    
    async def resolve_async(self, next_: Callable[..., Awaitable[Any]], source: Any, info: Info, **kwargs: Any) -> Any:
        
        
        """ Resolve the field with authentication checks."""
        if not info.context.request.user:
            raise GraphQLError("Authentication required")
        
        
        try:
            token: JWTToken = info.context.request.get_extension("token")
            
            if self.scopes and not token.has_scopes(self.scopes):
                raise GraphQLError(f"User does not have the required scopes: {self.scopes}")
        
            if self.roles and not token.has_roles(self.roles):
                raise GraphQLError(f"User does not have the required roles: {', '.join(self.roles)}")
            
            
        except KeyError:
            raise GraphQLError("Token not found in request context")
        
        
        
        
        return await next_(source, info, **kwargs)
  
  
class AuthSubscribeExtension(FieldExtension):
    
    
    def __init__(self, scopes: Optional[List[str]] | str = None, roles: Optional[List[str]] = None) -> None:
        """Initialize the AuthExtension with optional scopes and roles."""
        if isinstance(scopes, str):
            scopes = [scopes]
        if roles and isinstance(roles, str):
            roles = [roles]
        
        self.scopes: Optional[List[str]] = scopes
        self.roles: Optional[List[str]] = roles
        
        
        
    def apply(self, field: StrawberryField) -> None:
        """Apply the Auth directive to the field.

        Args:
            field (StrawberryField): The authentication field to which the directive will be applied.
        """
        assert field.is_subscription, "AuthSubscribeExtension can only be applied to subscription fields."
        field.directives.append(Auth(required_scopes=self.scopes, required_roles=self.roles))

    def resolve(
        self, next_: Callable[..., Any], source: Any, info: Info, **kwargs
    ) -> Any:
        """ Resolve the field with authentication checks."""
        if not info.context.request.user:
            raise GraphQLError("Authentication required")
        
        
        try:
            token: JWTToken = info.context.request.get_extension("token")
            
            if self.scopes and not token.has_scopes(self.scopes):
                raise GraphQLError(f"User does not have the required scopes: {self.scopes}")
        
            if self.roles and not token.has_roles(self.roles):
                raise GraphQLError(f"User does not have the required roles: {', '.join(self.roles)}")
            
            
        except KeyError:
            raise GraphQLError("Token not found in request context")
        
        
        
        return next_(source, info, **kwargs)
    
    async def resolve_async(self, next_: Callable[..., Awaitable[Any]], source: Any, info: Info, **kwargs: Any) -> Any:
        
        
        """ Resolve the field with authentication checks."""
        if not info.context.request.user:
            raise GraphQLError("Authentication required")
        
        
        try:
            token: JWTToken = info.context.request.get_extension("token")
            
            if self.scopes and not token.has_scopes(self.scopes):
                raise GraphQLError(f"User does not have the required scopes: {self.scopes}")
        
            if self.roles and not token.has_roles(self.roles):
                raise GraphQLError(f"User does not have the required roles: {', '.join(self.roles)}")
            
            
        except KeyError:
            raise GraphQLError("Token not found in request context")
        
        
        
        # this is a workaround for the fact that strawberry does not support async resolvers for subscriptions
        return next_(source, info, **kwargs)  
    
    
    
    
all_directives = [Auth]


""" Authentikate is a simple token based authentication library for Python

Authentikate is a simple token based authentication library for Python. It 
provides convenient functions to authenticate requests and decode tokens, 
authenticating users and applications.

It is designed to be used with Django, but can be used with any Python
framework.

Supported Token Types
- JWT (JSON Web Tokens) (with client_id, user_id, scopes, and expiration)
- Static tokens (for testing and pre-defined tokens)


 """
